@@ -15,29 +15,51 @@ IPTOS_MINCOST = 0x02
 DEFAULT_LIMIT = 2 ** 20  # 1Mb
 
 
-log = logging.getLogger('sockio')
+log = logging.getLogger("sockio")
 
 
 def ensure_connection(f):
     assert asyncio.iscoroutinefunction(f)
+
     @functools.wraps(f)
     async def wrapper(self, *args, **kwargs):
-        if self.auto_reconnect and not self.connected:
+        if self.auto_reconnect and not self.connected():
             await self.open()
-        return await f(self, *args, **kwargs)
+        timeout = kwargs.pop('timeout', self.timeout)
+        coro = f(self, *args, **kwargs)
+        if timeout is not None:
+            coro = asyncio.wait_for(coro, timeout)
+        return await coro
+
+    return wrapper
+
+
+def raw_handle_read(f):
+    assert asyncio.iscoroutinefunction(f)
+
+    @functools.wraps(f)
+    async def wrapper(self, *args, **kwargs):
+        try:
+            reply = await f(self, *args, **kwargs)
+        except ConnectionError:
+            await self.close()
+            raise
+        if not reply and self.at_eof():
+            await self.close()
+        return reply
+
     return wrapper
 
 
 class StreamReaderProtocol(asyncio.StreamReaderProtocol):
-
     def connection_lost(self, exc):
         result = super().connection_lost(exc)
-        self._exec_callback('connection_lost_cb', exc)
+        self._exec_callback("connection_lost_cb", exc)
         return result
 
     def eof_received(self):
         result = super().eof_received()
-        self._exec_callback('eof_received_cb')
+        self._exec_callback("eof_received_cb")
         return result
 
     def _exec_callback(self, name, *args, **kwargs):
@@ -49,13 +71,11 @@ class StreamReaderProtocol(asyncio.StreamReaderProtocol):
             if asyncio.iscoroutine(res):
                 self._loop.create_task(res)
         except Exception:
-            log.exception(
-                'Error in %s callback %r', name, callback.__name__)
+            log.exception("Error in %s callback %r", name, callback.__name__)
 
 
 class StreamReader(asyncio.StreamReader):
-
-    async def readline(self, eol=b'\n'):
+    async def readline(self, eol=b"\n"):
         # This implementation is a copy of the asyncio.StreamReader.readline()
         # with the purpose of supporting different EOL characters.
         # we walk on thin ice here: we rely on the internal _buffer and
@@ -80,12 +100,17 @@ class StreamReader(asyncio.StreamReader):
         self._buffer.clear()
 
 
-async def open_connection(host=None, port=None, loop=None,
-                          limit=DEFAULT_LIMIT, flags=0,
-                          on_connection_lost=None,
-                          on_eof_received=None,
-                          no_delay=True,
-                          tos=IPTOS_LOWDELAY):
+async def open_connection(
+    host=None,
+    port=None,
+    loop=None,
+    limit=DEFAULT_LIMIT,
+    flags=0,
+    on_connection_lost=None,
+    on_eof_received=None,
+    no_delay=True,
+    tos=IPTOS_LOWDELAY,
+):
     if loop is None:
         loop = asyncio.get_event_loop()
     reader = StreamReader(limit=limit, loop=loop)
@@ -93,22 +118,32 @@ async def open_connection(host=None, port=None, loop=None,
     protocol.connection_lost_cb = on_connection_lost
     protocol.eof_received_cb = on_eof_received
     transport, _ = await loop.create_connection(
-        lambda: protocol, host, port, flags=flags)
+        lambda: protocol, host, port, flags=flags
+    )
     writer = asyncio.StreamWriter(transport, protocol, reader, loop)
-    sock = writer.transport.get_extra_info('socket')
-    if hasattr(socket, 'TCP_NODELAY'):
-        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1 if no_delay else 0)
-    if hasattr(socket, 'IP_TOS'):
+    sock = writer.transport.get_extra_info("socket")
+    if hasattr(socket, "TCP_NODELAY") and no_delay:
+        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
+    if hasattr(socket, "IP_TOS"):
         sock.setsockopt(socket.SOL_IP, socket.IP_TOS, tos)
     return reader, writer
 
 
 class TCP:
-
-    def __init__(self, host, port, eol=b'\n', auto_reconnect=True,
-                 on_connection_made=None, on_connection_lost=None,
-                 on_eof_received=None, buffer_size=DEFAULT_LIMIT,
-                 no_delay=True, tos=IPTOS_LOWDELAY):
+    def __init__(
+        self,
+        host,
+        port,
+        eol=b"\n",
+        auto_reconnect=True,
+        on_connection_made=None,
+        on_connection_lost=None,
+        on_eof_received=None,
+        buffer_size=DEFAULT_LIMIT,
+        no_delay=True,
+        tos=IPTOS_LOWDELAY,
+        timeout=None
+    ):
         self.host = host
         self.port = port
         self.eol = eol
@@ -120,9 +155,10 @@ class TCP:
         self.on_eof_received = on_eof_received
         self.no_delay = no_delay
         self.tos = tos
+        self.timeout = timeout
         self.reader = None
         self.writer = None
-        self._log = log.getChild('TCP({}:{})'.format(host, port))
+        self._log = log.getChild("TCP({}:{})".format(host, port))
 
     def __del__(self):
         if self.writer is not None:
@@ -130,7 +166,7 @@ class TCP:
             if loop is not None and not loop.is_closed():
                 self.writer.close()
             else:
-                self._log.warning('could not close stream: loop closed')
+                self._log.warning("could not close stream: loop closed")
 
     def __aiter__(self):
         return self
@@ -138,19 +174,29 @@ class TCP:
     @ensure_connection
     async def __anext__(self):
         val = await self.reader.readline()
-        if val == b'':
+        if val == b"":
             raise StopAsyncIteration
         return val
 
-    async def open(self):
-        if self.connected:
-            raise ConnectionError('socket already open')
-        self._log.debug('open connection (#%d)', self.connection_counter + 1)
-        self.reader, self.writer = await open_connection(
-            self.host, self.port, limit=self.buffer_size,
+    async def open(self, **kwargs):
+        timeout = kwargs.get('timeout', self.timeout)
+        if self.connected():
+            raise ConnectionError("socket already open")
+        self._log.debug("open connection (#%d)", self.connection_counter + 1)
+        # make sure everything is clean before creating a new connection
+        await self.close()
+        coro = open_connection(
+            self.host,
+            self.port,
+            limit=self.buffer_size,
             on_connection_lost=self.on_connection_lost,
             on_eof_received=self.on_eof_received,
-            no_delay=self.no_delay, tos=self.tos)
+            no_delay=self.no_delay,
+            tos=self.tos,
+        )
+        if timeout is not None:
+            coro = asyncio.wait_for(coro, timeout)
+        self.reader, self.writer = await coro
         if self.on_connection_made is not None:
             try:
                 res = self.on_connection_made()
@@ -158,8 +204,9 @@ class TCP:
                     await res
             except Exception:
                 log.exception(
-                    'Error in connection_made callback %r',
-                    self.on_connection_made.__name__)
+                    "Error in connection_made callback %r",
+                    self.on_connection_made.__name__,
+                )
         self.connection_counter += 1
 
     async def close(self):
@@ -170,37 +217,65 @@ class TCP:
         self.reader = None
         self.writer = None
 
-    @property
     def in_waiting(self):
-        return len(self.reader) if self.connected else 0
+        return len(self.reader) if self.connected() else 0
 
-    @property
     def connected(self):
-        if self.reader is None:
-            return False
-        eof = self.reader.at_eof()
-        return not eof
+        return self.reader is not None and not self.at_eof()
 
+    def at_eof(self):
+        return self.reader is not None and self.reader.at_eof()
+
+    @raw_handle_read
     async def _read(self, n=-1):
         return await self.reader.read(n)
 
+    @raw_handle_read
+    async def _readexactly(self, n):
+        return await self.reader.readexactly(n)
+
+    @raw_handle_read
+    async def _readuntil(self, separator=b"\n"):
+        return await self.reader.readuntil(separator)
+
     async def _readline(self, eol=None):
         if eol is None:
             eol = self.eol
-        return await self.reader.readline(eol=eol)
+        try:
+            reply = await self.reader.readline(eol=eol)
+        except ConnectionError as err:
+            await self.close()
+            raise err
+        if not reply:
+            await self.close()
+        return reply
 
     async def _readlines(self, n, eol=None):
         if eol is None:
             eol = self.eol
-        return [await self.reader.readline(eol=eol) for _ in range(n)]
+        replies = []
+        for i in range(n):
+            reply = await self._readline(eol=eol)
+            if not reply:
+                break
+            replies.append(reply)
+        return replies
 
     async def _write(self, data):
-        self.writer.write(data)
-        await self.writer.drain()
+        try:
+            self.writer.write(data)
+            await self.writer.drain()
+        except ConnectionError:
+            await self.close()
+            raise
 
     async def _writelines(self, lines):
-        self.writer.writelines(lines)
-        await self.writer.drain()
+        try:
+            self.writer.writelines(lines)
+            await self.writer.drain()
+        except ConnectionError:
+            await self.close()
+            raise
 
     @ensure_connection
     async def read(self, n=-1):
@@ -216,17 +291,17 @@ class TCP:
 
     @ensure_connection
     async def readexactly(self, n):
-        return await self.reader.readexactly(n)
+        return await self._readexactly(n)
 
     @ensure_connection
-    async def readuntil(self, separator=b'\n'):
-        return await self.reader.readuntil(separator)
+    async def readuntil(self, separator=b"\n"):
+        return await self._readuntil(separator)
 
     @ensure_connection
     async def readbuffer(self):
         """Read all bytes currently available in the underlying buffer"""
-        size = self.in_waiting
-        return (await self._read(size)) if size else b''
+        size = self.in_waiting()
+        return (await self._read(size)) if size else b""
 
     @ensure_connection
     async def write(self, data):
@@ -254,27 +329,27 @@ class TCP:
         return await self._readlines(n, eol=eol)
 
     def reset_input_buffer(self):
-        if self.connected:
+        if self.connected():
             self.reader.reset()
 
 
 def parse_args(args=None):
     import argparse
+
     parser = argparse.ArgumentParser()
     log_level_choices = ["critical", "error", "warning", "info", "debug"]
     log_level_choices += [i.upper() for i in log_level_choices]
-    parser.add_argument('--host', default='0',
-                        help='SCPI device host name / IP')
-    parser.add_argument('-p', '--port', type=int, help='SCPI device port')
-    parser.add_argument('-r', '--request', default='*IDN?\n',
-                        help='SCPI request [%(default)s]')
-    parser.add_argument("--log-level", choices=log_level_choices,
-                        default="warning")
-    parser.add_argument('-d', '--debug', action='store_true')
+    parser.add_argument("--host", default="0", help="SCPI device host name / IP")
+    parser.add_argument("-p", "--port", type=int, help="SCPI device port")
+    parser.add_argument(
+        "-r", "--request", default="*IDN?\n", help="SCPI request [%(default)s]"
+    )
+    parser.add_argument("--log-level", choices=log_level_choices, default="warning")
+    parser.add_argument("-d", "--debug", action="store_true")
     options = parser.parse_args(args)
-    if not options.request.endswith('\n'):
-        options.request += '\n'
-    fmt = '%(asctime)-15s %(levelname)-5s %(threadName)s %(name)s: %(message)s'
+    if not options.request.endswith("\n"):
+        options.request += "\n"
+    fmt = "%(asctime)-15s %(levelname)-5s %(threadName)s %(name)s: %(message)s"
     logging.basicConfig(level=options.log_level.upper(), format=fmt)
     return options
 
@@ -282,7 +357,7 @@ def parse_args(args=None):
 async def run(options):
     sock = TCP(options.host, options.port)
     request = options.request
-    nb_lines = request.count('\n')
+    nb_lines = request.count("\n")
     lines = await sock.write_readlines(request.encode(), nb_lines)
     for line in lines:
         print(line)
@@ -293,5 +368,5 @@ def main(args=None):
     return run(options)
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     asyncio.run(main())
